package cz.muni.fi.pb162.files;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.*;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.FileTime;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;

/**
 *
 */
public class DemoNIO {

    public static void main(String[] args) throws IOException {
        demoPath(); // Path interfece which is new in Java 7
        demoReadFileJ11(); // using Java 11 features
        demoFiles(); // using Java 7+ Files and Paths utility classes
        demoTraversal(); // traverse (visit) directories
        demoBufferedOps(); // convenient reading/writing text to file
        demoStreams(); // Java 8 Streams demo
    }

    /**
     * Path - An object that may be used to locate a file in a file system.
     */
    public static void demoPath() {
        System.out.println("\n*** Let's demonstrate Path class ***");
        // path to subfolder relative to current dir
        // Path.of works in Java 11+ otherwise Path is available since Java 7

        // identify folder/subfolder
        Path subfldr = Path.of("folder", "subfolder");

        // identify file in the subfolder
        Path file = subfldr.resolve("file.txt");

        // simply identify a sibling file (ie. in same directory/folder)
        Path siblingFile = file.resolveSibling("secondfile.txt");

        System.out.println("Subfolder path resolves to: " + subfldr);
        System.out.println("File path resolves to: " + file);
        System.out.println("Sibling file resolves to: " + siblingFile);

        // Iterate over path segments
        System.out.println("Sibling file path segments:");
        for (Path segment : siblingFile)
            System.out.print(" segment=" + segment);

        // crazy paths can also be resolved
        Path relative = Path.of("..", "subfolder");
        Path subfldrUgly = subfldr.resolve(relative);
        System.out.println("\nSubfolder path ugly: " + subfldrUgly);
        // normalize it
        Path subfldrNorm = subfldrUgly.normalize();
        System.out.println("Subfolder path normalized: " + subfldrNorm);
    }

    public static void demoReadFileJ11() throws IOException {
        System.out.println("\n*** Let's demonstrate reading entire file (Java 11) ***");
        // from https://www.techiedelight.com/read-contents-file-java-11/
        String s = readFile("text.txt");
        System.out.println("Content read from text.txt: " + s);
    }

    /**
     * Method to readString from file with UTF-8 encoding
     */
    private static String readFile(String filename) throws IOException {
        // Paths "utility class" in java.nio allows to create Path
        Path path = Paths.get(filename);
        // Files "utility class" in java.nio features readString method
        // always using UTF-8 as input file encoding
        return Files.readString(path);
    }

    /**
     * Method to readString from file with any encoding
     */
    public static String readFile(String path, Charset encoding) throws IOException {
        return Files.readString(Paths.get(path), encoding);
    }

    public static void demoFiles() throws IOException {
        System.out.println("\n*** Let's demonstrate Files & Paths utility class (Java 7+) ***");
        System.out.println("Files.copy file.txt -> dest.txt");
        Path src = Path.of("folder", "subfolder", "file.txt");
        Path dest = Path.of("folder", "subfolder", "dest.txt");

        System.out.println("Deleted dest.txt before copy? " + Files.deleteIfExists(dest));

        // let us copy src file to dest file with options
        // copy options can eg. force to overwrite destination if exists
        Files.copy(src, dest, StandardCopyOption.COPY_ATTRIBUTES);
        // Files.move works per analogy

        BasicFileAttributes a = Files.readAttributes(dest, BasicFileAttributes.class);
        System.out.println("Attributes of dest.txt: " + a);
        FileTime ftcr = a.creationTime(); // Returns the creation time.
        System.out.println("creationTime: " + ftcr);

        Object o = a.fileKey(); // Returns an object that uniquely identifies the given file, or null if a file
        // key is not available.
        System.out.println("fileKey: " + o);

        boolean dir = a.isDirectory(); // Tells whether the file is a directory.
        System.out.println("isDirectory? " + dir);

        boolean oth = a.isOther(); // Tells whether the file is something other than a regular file, directory, or
        // symbolic link.
        System.out.println("isOther? " + oth);

        boolean regf = a.isRegularFile(); // Tells whether the file is a regular file with opaque content.
        System.out.println("isRegularFile? " + regf);

        boolean sl = a.isSymbolicLink(); // Tells whether the file is a symbolic link.
        System.out.println("isSymbolicLink? " + sl);

        FileTime ftla = a.lastAccessTime(); // Returns the time of last access.
        System.out.println("lastAccessTime: " + ftla);

        FileTime ftlm = a.lastModifiedTime(); // Returns the time of last modification.
        System.out.println("lastModifiedTime: " + ftlm);
        System.out.println("lastModifiedTime.to(HOURS): " + ftlm.to(TimeUnit.HOURS));
        System.out.println("lastModifiedTime.toMillis(): " + ftlm.toMillis());

        long size = a.size(); // Returns the size of the file (in bytes).
        System.out.println("size: " + size);
    }

    public static void demoTraversal() throws IOException {
        System.out.println("\n*** Let's demonstrate visiting files in dir (Java 7+) ***");
        System.out.println("Files *.java and *.class in .");
        Path dir = Path.of(".");
        try (DirectoryStream<Path> stream = Files.newDirectoryStream(dir, "*.{java,class}")) {
            for (Path entry : stream)
                System.out.println(entry);
        } // we could possibly add catch block for errors

        System.out.println("Recursive walk through files in . (max depth 2)");
        try (Stream<Path> stream = Files.walk(dir, /* max depth */ 2)) {
            stream.forEach(System.out::println);
        } // we could possibly add catch block for errors

        /*
         * Recursive walk through files in . (max depth 2) . ./.upm ./.upm/store.json
         * ./Main.class ./folder ./folder/subfolder ./text.txt ./Main$1.class
         * ./Main.java
         */
        System.out.println("Recursive walk using FileVisitor");
        Files.walkFileTree(dir, new SimpleFileVisitor<>() {
            @Override
            public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) {
                System.out.println("preVisitDirectory " + dir);
                return FileVisitResult.CONTINUE;
            }

            @Override
            public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) {
                System.out.println("visitFile " + file);
                return FileVisitResult.CONTINUE;
            }

            @Override
            public FileVisitResult visitFileFailed(Path file, IOException exc) {
                System.out.println("visitFileFailed " + file);
                return FileVisitResult.CONTINUE;
            }

            @Override
            public FileVisitResult postVisitDirectory(Path dir, IOException exc) {
                System.out.println("postVisitDirectory " + dir);
                return FileVisitResult.CONTINUE;
            }
        });
    }

    public static void demoBufferedOps() throws IOException {
        System.out.println("\n*** Let's demonstrate buffered read/write (Java 8+) ***");
        // opens text file in UTF-8
        // same as Files.newBufferedReader(path, StandardCharsets.UTF_8)
        try (BufferedReader br = Files.newBufferedReader(Path.of("text.txt"))) {
            String s = br.readLine();
            int i = 1;
            System.out.println("Reading file ./text.txt");
            while (s != null) {
                System.out.println("Line " + i + ". " + s);
                s = br.readLine();
                i++;
            }
        }
    }

    public static void demoStreams() {
        System.out.println("\n*** Let's demonstrate streams (Java 8+) ***");

        // do not confuse with Input/OutputStreams!
        // this takes lines from file into stream
        try (Stream<String> stream = Files.lines(Path.of("text.txt"))) {
            // sumarizes 3rd fields (as integers)
            int sum = stream
                    .filter(l -> l.startsWith("H")) // take only those H* greetings
                    .mapToInt(l -> Integer.parseInt(l.split(" ")[2])) // take third field (number)
                    .sum(); // sum the numbers
            System.out.println("Sum of greetings H* " + sum);
        } catch (Exception e) {
            System.out.println("Exception when reading file " + e);
        }
    }
}
