package cz.muni.fi.pb162.iostreams;

import java.io.*;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;

/**
 * Vstupno-výstupné operácie sú v Jave založené na streamoch
 * Keďže všetky sú založené na streamoch, v implementácii úplne nezáleží o aké V/V zariadenie ide
 * Súčastne sú vďaka tomu V/V operácie platformovo nezávislé
 *
 * V tomto deme si ukážeme V/V operácie nad textovými dátami
 */
public class Demo {
    public static void main(String[] args) throws IOException {

        File input = new File("streamFiles", "input.txt");

        // prečítame všetky riadky zo súboru a vrátime len tie, ktoré nie sú prázdne
        List<String> lines = readLinesWithBufferedReader(input);

        // ak súbor, ktorý si vyberieme ako output neexistuje, tak sa vytvorí
        // ak existuje jeho obsah je vymazány pred zápisom
        File output = new File("streamFiles", "output.txt");

        // zapíšeme všetky riadky do výstupného súboru
        writeLinesIntoFile(lines, output);

        File inputText = new File("streamFiles", "inputText.txt");
        // prečítame všetky vety(po znak ukončenia vety "./!/?") v súbore a vrátime ich ako zoznam Stringov
        List<String> textLines = readSentencesWithScanner(inputText);

        File outputText = new File("streamFiles", "outputText.txt");
        // zapíšeme všetky načítané vety na samostatný riadok
        writeLinesIntoFile(textLines, outputText);

    }

    /**
     * Ukážka využitia BufferReaderu pri čítaní zo súboru
     */
    private static List<String> readLinesWithBufferedReader(File input) throws IOException {
        List<String> lines = new ArrayList<>();

        // používame try-with-resources
        // try-with-resources zabezpečí, že ak dôjde k výnimke tak sa vstupný alebo výstupný súbor zavrie
        // náhrada za starší try-catch-finally kde sa v bloku finally volalo close()
        // InputStream je binárny vstupný prúd, ktorý umožňuje čítať binárne dáta
        try (InputStream is = new FileInputStream(input)) {
            // InputStreamReader je znakový vstupný prúd, umožňuje čítať po znakoch do bufferu
            InputStreamReader reader = new InputStreamReader(is); // UTF-8
            // BufferedReader umožňuje čítať po riadkoch a preto je najvhodnejší pre naše použitie
            BufferedReader br = new BufferedReader(reader);
            while (br.ready()) {
                String line = br.readLine();
                String[] fields = line.split("\\:");
                // zahodíme prázdne riadky
                if (line.length() > 0) {
                    lines.add(line);
                }
            }
        }
        return lines;
    }

    /**
     * Ukážka použitia BufferedWriter na zápis do súboru
     */
    private static void writeLinesIntoFile(List<String> lines, File output) throws IOException {
        // používame try-with-resources
        // OutputStream slúži na zápis binárných dát
        try (OutputStream os = new FileOutputStream(output)) {
            // OutputStreamWriter slúži na zápis znakov
            OutputStreamWriter writer = new OutputStreamWriter(os);
            // BufferedWriter je veľmi podobný OutputStreamWriteru avšak obsahuje naviac metódu newLine()
            // tá slúži na vloženie znaku nového riadku
            BufferedWriter bw = new BufferedWriter(writer);
            for (String line : lines) {
                bw.write(line);
                bw.newLine();
            }
            // zapíše dáta zo streamu do súboru
            // po zakomentovaní by ostal výstupný súbor prázdny
            bw.flush();
        }
    }

    /**
     * Ukážka využitia Scanneru
     */
    private static List<String> readSentencesWithScanner(File input) throws IOException {
        List<String> lines = new ArrayList<>();
        // Scanner je jeden z najjednoduchších spôsobov ako načítať vstup
        try (Scanner scanner = new Scanner(input)) {
            // pattern určuje aký formát majú mať dáta, ktoré chceme načítať
            String pattern = "[^.!?]*[.?!]\\s*";
            while (scanner.hasNext()) {
                // načíta zo vstupu dáta, ktoré spĺňajú formát
                String line = scanner.findInLine(pattern);
                lines.add(line);
            }
        }
        return lines;
    }
}
